/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.kelp.common.wizard;

import java.util.StringTokenizer;

import org.enhydra.kelp.common.node.OtterProject;
import org.enhydra.kelp.common.node.OtterSelectionNode;
import org.enhydra.kelp.common.capture.Writer;
import org.enhydra.kelp.common.capture.WriteListener;

import java.io.PrintWriter;
import java.io.File;

/**
 * CompileThread calls XMLC once for each selected HTML file.
 *
 *
 * @author
 * Paul Mahar
 * @version %I%, %G%
 */
abstract public class AbstractBuildThread extends Thread {
    private boolean       printProgress = true;
    private boolean       fresh = true;
    private OtterProject  project = null;
    private OtterSelectionNode[]  nodes = null;
    private PrintWriter      traceWriter = null;

    /**
     * Create a CompileThread with listeners ready to
     * recieve output from XMLC.
     *
     *
     * @param listeners
     * A listener implements an onWrite method that is called
     * whenever XMLC writes output. The writes are captured
     * for display in a wizard.
     */

    public AbstractBuildThread() {
      this(new WriteListener[0]);
    }

    public AbstractBuildThread(WriteListener[] listeners) {
      this(listeners, false);
    }

    public AbstractBuildThread(WriteListener[] listeners,
                         boolean buffer) {
	Writer capture = new Writer();
        capture.setLineBuffered(buffer);

	for (int i = 0; i < listeners.length; i++) {
	    capture.addWriteListener(listeners[i]);
	}

	traceWriter = new PrintWriter(capture, false);
  }



   /**
     * Loop through all selected nodes files and build them.
     * Write out exceptions to any registered listeners.
     */
    abstract public void make();

    /**
     * Determine what files were created and write out the
     * build results to any registered listeners.
     */
    abstract public void printResults();

    /**
     * Get the HTML source files that are to be compiled.
     */
    public OtterSelectionNode[] getNodes() {
	return nodes;
    }

    /**
     * Set the HTML source files to compile.
     */
    public void setNodes(OtterSelectionNode[] n) {
	nodes = n;
    }

    /**
     * Get the IDE project that contains the HTML source files to compile.
     */
    public OtterProject getProject() {
	return project;
    }

    /**
     * Set the IDE project that contains the HTML source files to compile.
     */
    public void setProject(OtterProject bp) {
	project = bp;
    }

    public boolean isPrintProgress() {
      return printProgress;
    }

    public void setPrintProgress(boolean b) {
      printProgress = b;
    }

    public PrintWriter getTraceWriter() {
      return traceWriter;
    }



    /**
     * Run the XMLC for all selected HTML files.
     */
    public void run() {
	fresh = true;
	make();
    }

    public synchronized void runInCurrentThread() {
	fresh = true;
	make();
    }

    /**
     * Make the thread stale so it stops running gracefully.
     */
    public void expire() {
	fresh = false;
    }

    public boolean isFresh() {
      return fresh;
    }


    public void printExceptions() {
        StringTokenizer tokenizer = null;
      	int eCount = 0;
        int totalCount = getNodes().length;
      	for (int i = 0; i < totalCount; i++) {
	    if (getNodes()[i].isSelected() &&
                getNodes()[i].getException() != null) {
		eCount++;
	    }
	}
	if (eCount > 0) {
	    traceWriter.println("");
	    traceWriter.println("Exception encountered:");

	    for (int i = 0; i < totalCount; i++) {
		if (getNodes()[i].getException() != null) {
		    traceWriter.print("    ");
		    traceWriter.print(getNodes()[i].getFilePath());
		    traceWriter.println(":");
                    tokenizer  = new StringTokenizer(getNodes()[i].getException().getMessage(),"\n");
                    while (tokenizer.hasMoreTokens()) {
                      traceWriter.println(tokenizer.nextToken());
                    }
                }
	    }
	}
    }

    /**
     * Sleep the thread for a given number of seconds.
     *
     *
     * @param seconds
     * How many seconds to sleep.
     */
    public void sleepInSeconds(double seconds) {
	try {
	    Thread.sleep((int) (seconds * 1000));
	} catch (InterruptedException e) {}
    }

}
