/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */
package org.enhydra.kelp.common.wizard.xmlc;

// Kelp imports
import org.enhydra.kelp.common.bridge.Generator;
import org.enhydra.kelp.common.bridge.Parser;
import org.enhydra.kelp.common.bridge.XMLCConnectionFactory;
import org.enhydra.kelp.common.node.OtterXMLCNode;

// Enhydra imports
import org.enhydra.xml.xmlc.commands.xmlc.*;
import org.enhydra.xml.xmlc.commands.options.*;
import org.enhydra.xml.xmlc.compiler.EditDOM;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.parsers.*;
import org.enhydra.xml.xmlc.XMLCUtil;
import org.enhydra.xml.xmlc.XMLCVersion;
import org.enhydra.xml.xmlc.XMLCException;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

// Standard imports
import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Class derived from the XMLC compiler program for use in this wizard.
 * TODO: Mark may change some things that will require changes to this class.
 */
public class XMLCRunner {

    /*
     * Object containing the parsed standard options.
     */
    private OtterXMLCNode node;

    /*
     * Document being compiled.
     */
    private Document      document;

    /*
     * Trace and verbose output.
     */
    private PrintWriter   traceWriter = new PrintWriter(System.err, true);

    /*
     * XML DOM management object.
     */
    private XMLCDocument  xmlcDoc;

    /**
     * Constructor declaration
     *
     *
     * @see
     */
    public XMLCRunner() {}

    /**
     * Method declaration
     *
     *
     * @param tw
     *
     * @see
     */
    public void setTraceWriter(PrintWriter tw) {
        traceWriter = tw;
    }

    /**
     * Method declaration
     *
     *
     * @return
     *
     * @see
     */
    public PrintWriter getTraceWriter() {
        return traceWriter;
    }

    /*
     * Print out a stack trace.  This works around bugs in SAXException that
     * cause printing a stack trace to generate a null pointer exception.  We
     * can't just fix this in the OpenXML source (we did), because other
     * parsers include that same broken class.
     */

    /**
     * Method declaration
     *
     *
     * @param except
     *
     * @see
     */
    private void printStackTrace(Throwable except) {

        // FIXME: needed???
        if (except instanceof SAXException) {
            Exception except2 = ((SAXException) except).getException();

            if (except2 != null) {
                except2.printStackTrace();
                return;
            }
        }
        except.printStackTrace();
    }

    /*
     * Handle an exception, printing the approriate error
     * message.
     */

    /**
     * Method declaration
     *
     *
     * @param except
     *
     * @see
     */
    public void printException(Throwable except) {

        // Must check that options have been parsed before using them.
        if (except instanceof XMLCException) {
            System.err.println("Error: " + except.getMessage());
            if ((node.getMetaDataHandler() != null)
                    && (node.getMetaDataHandler().getVerbose())) {
                printStackTrace(except);
            }

            // If this is result of a java.lang exception, print stack.
            Throwable cause = ((XMLCException) except).getCause();

            if ((cause != null)
                    && cause.getClass().getName().startsWith("java.lang.")) {
                printStackTrace(cause);
            }
        } else if ((except instanceof FileNotFoundException)
                   || (except instanceof IOException)) {
            System.err.println("Error: " + except.getClass().getName() + ": "
                               + except.getMessage());
            if ((node.getMetaDataHandler() != null)
                    && (node.getMetaDataHandler().getVerbose())) {
                printStackTrace(except);
            }
        } else {
            System.err.println("Error: " + except.getClass().getName() + ": "
                               + except.getMessage());
            printStackTrace(except);
        }
        System.exit(1);
    }

    /*
     * Parse the page into the DOM and perform various checks and edits.
     */

    /**
     * Method declaration
     *
     *
     * @throws FileNotFoundException
     * @throws IOException
     * @throws SAXException
     * @throws XMLCException
     *
     * @see
     */

    /*
     * Parse the page into the DOM and perform various checks and edits.
     */
    private void parsePage()
            throws XMLCException, FileNotFoundException, IOException {
        if (node.getMetaDataHandler().getVerbose()) {
            traceWriter.println(">>> parsing "
                                + node.getMetaDataHandler().getInputDocument());
        }
        Parser  parser = null;
        PrintWriter infoWriter = null;
        if (node.getMetaDataHandler().getPrintParseInfo()) {
          infoWriter = traceWriter;
        }

        parser = XMLCConnectionFactory.createParser(traceWriter, infoWriter);
        parser.setMetaDataHandler(node.getMetaDataHandler());

        xmlcDoc = (XMLCDocument) parser.parse();
        document = xmlcDoc.getDocument();
        if (node.getMetaDataHandler().getPrintParseInfo()) {
            if (node.getMetaDataHandler().getVerbose()) {
                traceWriter.println(">>> printing document info");
            }
            LocalPrintInfo info = new LocalPrintInfo(document, xmlcDoc);

            info.printInfo(traceWriter);
        }
        EditDOM editDOM = null;
        editDOM = XMLCConnectionFactory.createEditDOM(node.getMetaDataHandler());
        editDOM.edit(xmlcDoc);

        if (node.getMetaDataHandler().getPrintDOM()) {
            if (node.getMetaDataHandler().getVerbose()) {
                traceWriter.println(">>> dumping document hierarchy");
            }
            XMLCUtil.printNode("DOM hierarchy", document, XMLCUtil.PRINT_ALL,
                               traceWriter);
        }
    }

    /*
     * Generate the Java source files.
     */

    /**
     * Method declaration
     *
     *
     * @exception IOException
     * @exception XMLCException
     *
     * @see
     */
    private void generateJavaSource() throws XMLCException, IOException {
        PrintWriter accessorWriter = null;
        PrintWriter verboseWriter = null;
        Generator codeGen = null;

        if (node.getMetaDataHandler().getPrintAccessorInfo()) {
          accessorWriter = traceWriter;
        }
        if (node.getMetaDataHandler().getVerbose()) {
            traceWriter.println(">>> generating code");
            verboseWriter = traceWriter;
        }
        codeGen = XMLCConnectionFactory.createGenerator(
            node.getMetaDataHandler(),
            xmlcDoc,
            accessorWriter);
        codeGen.generateJavaSource(verboseWriter);
           //        ClassGenerator codeGen =
//            new ClassGenerator(node.getMetaDataHandler().getMetaData(), xmlcDoc,
//                               (node.getMetaDataHandler().getPrintAccessorInfo() ? traceWriter : null));
//        codeGen.generateJavaSource(node.getMetaDataHandler().getVerbose() ? traceWriter : null);
    }

    /*
     * Generate method information without generating code.
     */

    /**
     * Method declaration
     *
     *
     * @exception IOException
     * @exception XMLCException
     *
     * @see
     */
    private void generateJavaSourceInfo() throws XMLCException, IOException {
        if (node.getMetaDataHandler().getVerbose()) {
            traceWriter.println(">>> generating code info without generating source");
        }
        Generator codeGen = XMLCConnectionFactory.createGenerator(
            node.getMetaDataHandler(), xmlcDoc,
           (node.getMetaDataHandler().getPrintAccessorInfo() ? traceWriter : null));
        codeGen.generateJavaSource(node.getMetaDataHandler().getVerbose() ? traceWriter : null);
    }

    /**
     * Write the DOM to a file.
     */
    private void writeDOM() throws IOException {
        File outputFile = new File(node.getMetaDataHandler().getDocumentOutput());
        String dir = outputFile.getParent();

        if (dir != null) {
            new File(dir).mkdirs();
        }
        PrintWriter domOut =
            new PrintWriter(new BufferedOutputStream(new FileOutputStream(node.getMetaDataHandler().getDocumentOutput())));

        try {
            domOut.println(xmlcDoc.toDocument());
        }
        finally {
            domOut.close();
        }
    }

    /*
     * Generate the Jave source and Compile the page.
     */

    /**
     * Method declaration
     *
     *
     * @exception FileNotFoundException
     * @exception IOException
     * @exception XMLCException
     *
     * @see
     */
    private void compilePage()
            throws XMLCException, FileNotFoundException, IOException {
        generateJavaSource();

        // do not compile Java and always check for recomp.
        if (node.getMetaDataHandler().getRecompilation()) {
            File op = node.getOptionFileForRecomp();
            if (! op.getParentFile().exists()) {
              op.getParentFile().mkdirs();
            }
            node.getMetaDataHandler().save(node.getOptionFileForRecomp());
//            node.getMetaDataHandler().getMetaData().getMetaDataDocument().serialize();
        }

        // do not compile Java
        // never delete source.
    }

    /**
     * Print the XMLC version number.
     */
    private void printVersion() {
        System.out.println("Enhydra XMLC version " + XMLCVersion.VERSION);
        System.out.println("See http://www.enhydra.org for latest distribution");
    }

    /*
     * Parse arguments and compile the page.
     */

    /**
     * Method declaration
     *
     *
     * @throws FileNotFoundException
     * @throws IOException
     * @throws SAXException
     * @throws XMLCException
     *
     * @see
     */
    public void compile()
            throws XMLCException, FileNotFoundException, IOException,
                   SAXException {

        // Setup and document parsing.
        node.initProjectOptions();
        if (node.getMetaDataHandler().getPrintVersion()) {
            printVersion();
            if (node.getMetaDataHandler().getInputDocument() == null) {
                return;
            }
        }
        parsePage();
        if (node.getMetaDataHandler().getDocumentOutput() == null) {
            compilePage();
        } else {
            if (node.getMetaDataHandler().getPrintAccessorInfo()) {
                generateJavaSourceInfo();
            }
            writeDOM();
        }
        if (node.getMetaDataHandler().getVerbose()) {
            traceWriter.println(">>> completed");
        }
    }

    /**
     * Method declaration
     *
     *
     * @param cn
     *
     * @see
     */
    public void setNode(OtterXMLCNode n) {
        node = n;
    }

    /**
     * Class declaration
     *
     *
     * @author Paul Mahar
     * @version $Revision$
     */
    private class LocalPrintInfo {

        /**
         * Table URLs.
         */
        private Vector urls = new Vector();

        /**
         * Table of element ids.
         */
        private Vector ids = new Vector();

        /**
         * Construct document info from document.
         */
        public LocalPrintInfo(Document doc, XMLCDocument xmlcDoc) {
            getNodeInfo(doc, xmlcDoc);
        }

        /**
         * Append URLs from an element.
         */
        private void getElementURLs(Element element, XMLCDocument xmlcDoc) {
            String[] attrs = xmlcDoc.getElementURLAttrs(element);

            if (attrs == null) {
                return;    // No URLs in this node.
            }
            for (int idx = 0; idx < attrs.length; idx++) {
                String url = element.getAttribute(attrs[idx]);

                if ((url != null) && (url.length() > 0)) {
                    urls.addElement(url);
                }
            }
        }

        /**
         * Recursively scan nodes nodes and accumulate information.
         */
        private void getNodeInfo(Node node, XMLCDocument xmlcDoc) {
            if (node instanceof Element) {
                Element elem = (Element) node;

                getElementURLs(elem, xmlcDoc);
                String id = xmlcDoc.getElementId(elem);

                if ((id != null) && (id.length() > 0)) {
                    ids.addElement(id + " => "
                                   + xmlcDoc.nodeClassToInterface(node));
                }
            }
            for (Node child = node.getFirstChild(); child != null;
                    child = child.getNextSibling()) {
                getNodeInfo(child, xmlcDoc);
            }
        }

        /**
         * Print document information.
         */
        protected void printInfo(PrintWriter out) {
            int idx;

            out.println("Element IDs:");
            for (idx = 0; idx < ids.size(); idx++) {
                out.println("    " + ids.elementAt(idx));
            }
            out.println("Document URLs:");
            for (idx = 0; idx < urls.size(); idx++) {
                out.println("    " + urls.elementAt(idx));
            }
        }

    }
}
